'''Client for connecting with Gaia machines'''
import threading
import json
import requests
import websocket


class Client:
    '''Client for connecting with Gaia machines'''

    def __init__(
            self,
            address,
            user=None,
            pwd=None,
            machine_state_callback=None,
    ):

        # Threading event for waiting that the test box is started to close
        self.wait_closing_event = threading.Event()

        # Threading event for waiting that the test box is ready for testing
        self.wait_ready_event = threading.Event()

        # Threading event for waiting that the test box is not ready for testing
        self.wait_not_ready_event = threading.Event()

        self.state_lock = threading.RLock()

        if user and pwd:
            self.requests = requests.Session()

            self.requests.post(address + "/login", json={"user": user, "password": pwd})

        else:
            self.requests = requests

        def on_error(ws, error):
            '''Handle error'''
            print(error)

        def on_message(ws, message):
            '''Handle state change messages'''
            try:
                with self.state_lock:
                    message = json.loads(message)
                    if machine_state_callback:
                        machine_state_callback(message)

                    if message['state'] == 'Ready':
                        self.wait_ready_event.set()
                        self.wait_not_ready_event.clear()
                    else:
                        self.wait_ready_event.clear()
                        self.wait_not_ready_event.set()

                    if message['state'] == 'Closing' or message['state'] == 'Ready':
                        self.wait_closing_event.set()
                    else:
                        self.wait_closing_event.clear()

            except Exception as e:
                print(e)

        state_socket = websocket.WebSocketApp(
            "ws://" + address.strip("http://") + "/websocket/state", on_message=on_message
        )

        state_socket_thread = threading.Thread(target=state_socket.run_forever)
        state_socket_thread.setDaemon(True)
        state_socket_thread.start()

        self._applications = {}
        self.address = address

        # Get applications
        applications_json = self.requests.get(self.address + '/api/applications').json()
        entities = self._get_entities(applications_json)

        for entity in entities:
            if entity['properties']['name'] in self._applications:
                if entity['properties']['alias']:
                    self._applications[entity['properties']['alias']] = {
                        'actions': self._get_actions(entity),
                        'properties': entity['properties'],
                    }
            else:
                self._applications[entity['properties']['name']] = {
                    'actions': self._get_actions(entity),
                    'properties': entity['properties'],
                }

        root_json = self.requests.get(self.address + '/api').json()

        self.state_triggers = self._get_actions(root_json)

    @property
    def state(self):
        '''Returns state of gaia machine'''
        return self.requests.get(self.address + '/api').json()['properties']['state']

    @property
    def properties(self):
        '''Returns properties of gaia machine'''
        return self.requests.get(self.address + '/api').json()['properties']

    @property
    def applications(self):
        '''Returns all available applications'''
        return self._applications

    @property
    def ready_for_testing(self):
        '''Returns true if test box is fully available for all tests'''

        return self.state == 'Ready'

    @property
    def test_box_closing(self):
        '''Returns true if test box is test box is closing

        When test box is closing some tests may be executed. Note that
        on this case test box is not RF or audio shielded. Also because
        of safety reasons robot is not powered'''
        return self.state == 'Closing'

    def wait_ready(self, timeout=None):
        """Waits that the tester is ready and available for all tests.
        Timeout on seconds. Returns True if there was no timeout."""

        with self.state_lock:
            if self.ready_for_testing:
                return True

            return self.wait_ready_event.wait(timeout)

    def wait_closing(self, timeout=None):
        """Waits that the tester is closing.
        Timeout on seconds. Returns True if there was no timeout."""

        with self.state_lock:
            if self.test_box_closing:
                return True

            return self.wait_closing_event.wait(timeout)

    def wait_not_ready(self, timeout=None):
        """Waits that the tester is not ready.
        Timeout on seconds. Returns True if there was no timeout."""

        with self.state_lock:
            if not self.ready_for_testing:
                return True

            return self.wait_not_ready_event.wait(timeout)

    def _get_entities(self, json):
        '''Fetch entities from Siren entry'''

        entities = []
        for i in json['entities']:
            entities.append(i)
        return entities

    def _get_actions(self, entity):

        actions = {}
        try:
            entity_details = self.requests.get(entity['href']).json()
        except Exception as e:
            print(entity)

        for action in entity_details['actions']:
            actions[action['name']] = self._get_fields(action)
        # Add also blocked actions
        if 'blocked_actions' in entity_details:
            for action in entity_details['blocked_actions']:
                actions[action['name']] = self._get_fields(action)

        return actions

    def _get_fields(self, action):
        if action['method'] == 'POST':

            def post_func(**kwargs):
                '''Post function'''

                # Fields thats value is defined in API. "Static" fields.
                fields = {}
                for field in action['fields']:
                    if 'value' in field:
                        fields[field['name']] = field['value']

                # User defined fields. "Variable" fields
                fields.update(kwargs)
                request = self.requests.post(
                    json=fields, url=action['href'], headers={'Content-type': action['type']}
                )
                # TODO: Handle error nicely
                request.raise_for_status()

            return post_func

        else:

            def get_func():
                '''Get function'''
                request = self.requests.get(
                    url=action['href'], headers={'Content-type': action['type']}
                )
                # TODO: Handle error nicely
                request.raise_for_status()
                return request

            return get_func



import time
import json
import gaiaclient

CLIENT = gaiaclient.Client(
    'http://172.23.225.75',
    # Callback for state changes
    # machine_state_callback=print,
)

# Get state of the tester
print("State: " + CLIENT.state)

# This is how you get properties of application.
# For example here we get current position of X-axle of main robot.
print(CLIENT.applications['MainRobot']['properties']['position']['x'])


# Print available applications and actions
class GaiaJsonEncoder(json.JSONEncoder):
    '''Encode json properly'''
    def default(self, obj):
        if callable(obj):
            return obj.__name__
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


print(json.dumps(CLIENT.applications, indent=4, sort_keys=True, cls=GaiaJsonEncoder))
print(json.dumps(CLIENT.state_triggers, indent=4, sort_keys=True, cls=GaiaJsonEncoder))


def print_with_timestamp(msg):
    from datetime import datetime
    now = datetime.now()
    now = now.strftime("%H:%M:%S")
    print(now + ": " + msg)


while True:
    # From here starts the actual test sequence

    # Fake operator action. Put DUT in.
    CLIENT.applications['dut']['actions']['force-presence-on']()  # <-- DONT't USE IN PRODUCTION

    # Step 1: We are waiting that the test box gets ready and operator puts DUT(s) in

    # Step 2: Operator did put the DUT(s) in. DUT(s) is locked and it is safe to attach
    # battery connector, USB etc.The test box is still closing so it is
    # not audio or RF shielded and robot actions are not allowed

    # Optionally wait the test box closing
    CLIENT.wait_closing()  # <-- Optional

    print_with_timestamp("Test box closing!")

    # Wait that the test box is fully closed and ready for testing
    CLIENT.wait_ready()

    # Step 3: Test box is fully closed and we are ready for actual testing.
    print_with_timestamp("Ready for testing!")

    # Execute the tests. Here's some examples.

    # Run robot movement. See GcodeExample.GCode for g-code example and also how
    # to define tool on g-code.
    # Note that for safety reason when g-code is modified it will run once with low speed and power.
    # So if you made mistake and robot collides it won't brake anything
    # CLIENT.applications["MainRobot"]['actions']["cnc_run"](plain_text=GcodeExample.GCode)

    # Push button on DUT with pusher
    # CLIENT.applications["SideButtonPusher"]['actions']["Push"]()

    # Optionally wait that pusher is on end position (detected by sensor)
    # CLIENT.applications["SideButtonPusher"].wait_state("Push")

    # Release pusher
    # CLIENT.applications["SideButtonPusher"]['actions']["Release"]()

    # Record audio
    # (Not implemented in python CLIENT - yet!) CLIENT.applications["WaveRecorder"]['actions']["record-wave"](new Dictionary<string, object> { { "time_s", 2 }, { "filename", "testrecord.wav" } });

    # Play audio
    # (Not implemented in python CLIENT - yet!) CLIENT.Applications["WavePlayer"]['actions']["play-wave"](new Dictionary<string, object> { { "filename", "sine_1000Hz_-3dBFS_3s.wav" } });

    # Step 4: Testing is ready and we release the DUT and give test result so that test box can indicate it to operator
    CLIENT.state_triggers["ReleasePass"]()
    # The test box must be not ready after the release
    # If we don't wait here we might start the new sequence before last one is even ended
    CLIENT.wait_not_ready()
    print_with_timestamp("Test box not ready!")

    # Fake operator action. Take DUT out.
    time.sleep(2)  # <-- DO NOT USE ON PRODUCTION!
    CLIENT.applications['dut']['actions']['force-presence-off']()  # <-- DO NOT USE ON PRODUCTION!



import setuptools

'''
IMPORTANT!
Use python 2 to run setuptools. Otherwise this will end up as an package
at PyPi and we want it to be a module.

Commands to publish
python setup.py sdist bdist_wheel
twine upload dist/*

'''

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gaiaclient",
    version="1.2.2",
    license='MIT License',
    author="JOT Automation Ltd.",
    author_email="rami.rahikkala@jotautomation.com",
    description="Client for JOT Automation gaia machines",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jotautomation/gaiapythonclient",
    py_modules=['gaiaclient'],
    install_requires=['wheel', 'requests', 'websocket-client'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
